"""
The xcookie module is a CLI tool for initializing and maintaining standardized
repo infastructure. In other words it is a Python project cookie cutter that
attempts to help keep existing modules up to date with latest infastructure
developments.

Currently this module is specialized towards Erotemic / PyUtils / Kitware
projects but the goal is to eventually generalize everything.

+------------------+----------------------------------------------+
| Read the docs    | https://xcookie.readthedocs.io               |
+------------------+----------------------------------------------+
| Github           | https://github.com/Erotemic/xcookie          |
+------------------+----------------------------------------------+
| Pypi             | https://pypi.org/project/xcookie             |
+------------------+----------------------------------------------+

"""
__version__ = '0.2.1'
